from __future__ import annotations

from typing import *  # type: ignore

import rio

from .component import Component

__all__ = [
    "Container",
]


@final
class Container(Component):
    """
    A component holding a single child.

    `Container` is a simple container which holds a single child component. It
    is useful for when you receive a component as child and wish to add
    additional layout attributes such as a margin.


    ## Attributes

    `content`: The component to place inside the container.


    ## Examples

    This minimal example will simply display a `container` with the text "Hello
    World!":

    ```python
    rio.Container(rio.Text("Hello World!"))
    ```

    `Container`s are commonly used to add layout attributes to a single child.
    You can easily achieve this by adding the content to the container and then
    adding the layout attributes:

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Container(
                rio.Text("Hello World!"),
                margin=2,
            )
    """

    content: rio.Component

    def build(self) -> rio.Component:
        return self.content



import asyncio
import cProfile
import itertools
import pstats
import time
from typing import Callable

import rio

# Other options for profiling:
#
# https://github.com/jiffyclub/snakeviz
# https://github.com/nvdv/vprof
# https://vmprof.readthedocs.org/
# https://github.com/kwlzn/pytracing
# https://pypi.org/project/line-profiler


class ComplexComponent(rio.Component):
    def build(self) -> rio.Component:
        return rio.Column(
            rio.Text("Complex component"),
            rio.Row(
                rio.Container(rio.Button("Button 1")),
                rio.Container(rio.Button("Button 2")),
                rio.Container(rio.Button("Button 3")),
            ),
            rio.Text("End of complex component"),
        )


class BenchmarkComponent(rio.Component):
    child_factory: Callable[[], rio.Component] = lambda: rio.Text("Starting...")

    def __post_init__(self):
        self.benchmark_start_time = time.monotonic()
        self._child_factory_iter = itertools.islice(
            itertools.cycle(
                [
                    lambda: ComplexComponent(),
                    lambda: ComplexComponent(),
                    lambda: ComplexComponent(),
                ]
            ),
            500,
        )

    def build(self) -> rio.Component:
        asyncio.create_task(self._change())

        elapsed_time = int(time.monotonic() - self.benchmark_start_time)
        return rio.Column(
            rio.Text("Benchmarking..."),
            rio.Text(f"Elapsed time: {elapsed_time} sec"),
            self.child_factory(),
        )

    async def _change(self):
        try:
            self.child_factory = next(self._child_factory_iter)
        except StopIteration:
            self.session.close()
        else:
            await self.force_refresh()


def main():
    # app = rio.App(
    #     build=BenchmarkComponent,
    # )
    # app.run_in_window()
    import rio.cli

    rio.cli.run(release=True)


with cProfile.Profile() as profile:
    main()

    stats = pstats.Stats(profile)


stats.sort_stats(pstats.SortKey.CUMULATIVE)
stats.print_stats(r"[/\\]rio[/\\]")



from __future__ import annotations

import colorsys
import math
from typing import *  # type: ignore

from typing_extensions import TypeAlias
from uniserde import Jsonable

import rio

from .self_serializing import SelfSerializing

__all__ = [
    "Color",
    "ColorSet",
]


@final
class Color(SelfSerializing):
    """
    A color, optionally with an opacity.

    The `Color` class does exactly what it says on the tin: It represents a
    single color. They're used throughout Rio to specify the color of text,
    fills, and more.

    `Color` supports a variety of color spaces:

    ```python
    # Color from RGB(A)
    Color.from_rgb(1.0, 0.0, 0.0, 1.0)

    # Color from hex
    Color.from_hex("#ff0000")

    # Color from HSV(A)
    Color.from_hsv(0.0, 1.0, 1.0, 1.0)

    # Color from greyscale
    Color.from_grey(0.5, 1.0)

    # Invalid: Don't call `Color` directly
    # Color(1.0, 0.0, 0.0, 1.0)  # Raises a `RuntimeError`
    ```

    Regardless of how the color was created, all of the color's components will
    be accessible as attributes. For example, you can access `color.red`, even
    if the color was created from HSV values.


    ## Attributes

    `BLACK`: A pure black color.

    `GREY`: A medium grey color.

    `WHITE`: A pure white color.

    `RED`: A pure red color.

    `GREEN`: A pure green color.

    `BLUE`: A pure blue color.

    `CYAN`: A pure cyan color.

    `MAGENTA`: A pure magenta color.

    `YELLOW`: A pure yellow color.

    `PINK`: A pure pink color.

    `PURPLE`: A pure purple color.

    `ORANGE`: A pure orange color.

    `BROWN`: A pure brown color.

    `TRANSPARENT`: A fully transparent color.
    """

    _red: float
    _green: float
    _blue: float
    _opacity: float

    def __init__(self):
        raise RuntimeError(
            "Don't call `Color` directly. Use `from_rgb()` and related methods instead."
        )

    @classmethod
    def from_rgb(
        cls,
        red: float = 1.0,
        green: float = 1.0,
        blue: float = 1.0,
        opacity: float = 1.0,
    ) -> "Color":
        """
        Creates a color from RGB(A) values.

        Creates a color from RGB(A) values. All values must be between `0.0` and
        `1.0`, inclusive.

        If no `opacity` is given, the color will be fully opaque.

        ## Parameters

        red: The red component of the color. `0.0` is no red, `1.0` is full
            red.

        green: The green component of the color. `0.0` is no green, `1.0`
            is full green.

        blue: The blue component of the color. `0.0` is no blue, `1.0` is
            full blue.

        opacity: The opacity of the color. `0.0` is fully transparent,
            `1.0` is fully opaque.

        ## Raises

        ValueError: If any of the values are outside of the range `0.0` to
            `1.0`.
        """

        if red < 0.0 or red > 1.0:
            raise ValueError(f"`red` must be between 0.0 and 1.0, not {red}")

        if green < 0.0 or green > 1.0:
            raise ValueError(
                f"`green` must be between 0.0 and 1.0, not {green}"
            )

        if blue < 0.0 or blue > 1.0:
            raise ValueError(f"`blue` must be between 0.0 and 1.0, not {blue}")

        if opacity < 0.0 or opacity > 1.0:
            raise ValueError(
                f"`opacity` must be between 0.0 and 1.0, not {opacity}"
            )

        self = object.__new__(cls)

        self._red = red
        self._green = green
        self._blue = blue
        self._opacity = opacity

        return self

    @classmethod
    def from_hex(cls, hex_color: str) -> "Color":
        """
        Parses a hex string into a color.

        Parses a hex string into a color. A variety of formats are supported:

        - `rgb`
        - `rgba`
        - `rrggbb`
        - `rrggbbaa`

        All values may optionally be prefixed with a `#`.

        ## Raises

        ValueError: If the string is not a valid hex color.
        """
        # Drop any leading `#` if present
        hex_color = hex_color.removeprefix("#")

        # Make sure the string is the correct length
        if len(hex_color) not in (3, 4, 6, 8):
            raise ValueError(
                "The hex string must be 3, 4, 6 or 8 characters long"
            )

        # Split the string into the individual components
        if len(hex_color) == 3:
            rh, gh, bh = hex_color
            ah = "f"
            max = 15
        elif len(hex_color) == 4:
            rh, gh, bh, ah = hex_color
            max = 15
        elif len(hex_color) == 6:
            rh, gh, bh = hex_color[0:2], hex_color[2:4], hex_color[4:6]
            ah = "ff"
            max = 255
        else:
            rh, gh, bh, ah = (
                hex_color[0:2],
                hex_color[2:4],
                hex_color[4:6],
                hex_color[6:8],
            )
            max = 255

        # Parse it
        return cls.from_rgb(
            red=int(rh, 16) / max,
            green=int(gh, 16) / max,
            blue=int(bh, 16) / max,
            opacity=int(ah, 16) / max,
        )

    @classmethod
    def from_hsv(
        cls,
        hue: float,
        saturation: float,
        value: float,
        opacity: float = 1.0,
    ) -> "Color":
        """
        Create a color from HSV(A) values.

        Creates a color from HSV(A) values. All values must be between `0.0` and
        `1.0`, inclusive.

        If no `opacity` is given, the color will be fully opaque.

        ## Parameters

        hue: The hue of the color. `0.0` is red, `0.33` is green, `0.66` is
            blue, and `1.0` is red again.

        saturation: The saturation of the color. `0.0` is no saturation,
            `1.0` is full saturation.

        value: The value of the color. `0.0` is black, `1.0` is full
            brightness.

        opacity: The opacity of the color. `0.0` is fully transparent,
            `1.0` is fully opaque.

        ## Raises

        ValueError: If any of the values are outside of the range `0.0` to
            `1.0`.
        """
        if hue < 0.0 or hue > 1.0:
            raise ValueError("`hue` must be between 0.0 and 1.0")

        if saturation < 0.0 or saturation > 1.0:
            raise ValueError("`saturation` must be between 0.0 and 1.0")

        if value < 0.0 or value > 1.0:
            raise ValueError("`value` must be between 0.0 and 1.0")

        # Opacity will be checked by `from_rgb`

        return cls.from_rgb(
            *colorsys.hsv_to_rgb(hue, saturation, value),
            opacity=opacity,
        )

    @classmethod
    def from_grey(cls, grey: float, opacity: float = 1.0) -> "Color":
        """
        Creates a greyscale color.

        Creates a grey color with the given intensity. A `grey` value of 0.0
        corresponds to black, and 1.0 to white.

        If no `opacity` is given, the color will be fully opaque.

        ## Parameters

        grey: The intensity of the grey color. `0.0` is black, `1.0` is
            white.

        opacity: The opacity of the color. `0.0` is fully transparent,
            `1.0` is fully opaque.

        ## Raises

        ValueError: If `grey` is outside of the range `0.0` to `1.0`.
        """
        if grey < 0.0 or grey > 1.0:
            raise ValueError("`grey` must be between 0.0 and 1.0")

        # Opacity will be checked by `from_rgb`

        return cls.from_rgb(grey, grey, grey, opacity)

    @classmethod
    def _from_material_argb(cls, argb: int) -> "Color":
        """
        Creates a new color from the argb value, as used by the
        `material-color-utilities-python` package.
        """

        return cls.from_rgb(
            red=((argb >> 16) & 0xFF) / 255,
            green=((argb >> 8) & 0xFF) / 255,
            blue=(argb & 0xFF) / 255,
            opacity=((argb >> 24) & 0xFF) / 255,
        )

    @property
    def _as_material_argb(self) -> int:
        """
        Returns the argb value, as used by the `material-color-utilities-python`
        package.
        """

        return (
            int(round(self.opacity * 255)) << 24
            | int(round(self.red * 255)) << 16
            | int(round(self.green * 255)) << 8
            | int(round(self.blue * 255))
        )

    @property
    def red(self) -> float:
        """
        The red component of the color.

        The red component of the color. `0.0` is no red, `1.0` is full red.
        """
        return self._red

    @property
    def green(self) -> float:
        """
        The green component of the color.

        The green component of the color. `0.0` is no green, `1.0` is full
        green.
        """
        return self._green

    @property
    def blue(self) -> float:
        """
        The blue component of the color.

        The blue component of the color. `0.0` is no blue, `1.0` is full blue.
        """
        return self._blue

    @property
    def opacity(self) -> float:
        """
        How opaque the color appears.

        The opacity of the color. `0.0` is fully transparent, `1.0` is fully
        opaque.
        """
        return self._opacity

    @property
    def rgb(self) -> tuple[float, float, float]:
        """
        The color as RGB values.

        The color as RGB values. Each value is between `0.0` and `1.0`,
        inclusive.
        """
        return (self._red, self._green, self._blue)

    @property
    def rgba(self) -> tuple[float, float, float, float]:
        """
        The color as RGBA values.

        The color as RGBA values. Each value is between `0.0` and `1.0`,
        inclusive.
        """
        return (self._red, self._green, self._blue, self._opacity)

    @property
    def hsv(self) -> tuple[float, float, float]:
        """
        The color as HSV values.

        The color as HSV values. Each value is between `0.0` and `1.0`,
        inclusive.
        """
        return colorsys.rgb_to_hsv(self._red, self._green, self._blue)

    @property
    def hue(self) -> float:
        """
        The hue of the color.

        The hue of the color. `0.0` is red, `0.33` is green, `0.66` is blue, and
        `1.0` is red again.
        """
        return self.hsv[0]

    @property
    def saturation(self) -> float:
        """
        The saturation of the color.

        The saturation of the color. `0.0` is no saturation, `1.0` is full
        saturation.
        """
        return self.hsv[1]

    @property
    def value(self) -> float:
        """
        The value of the color.

        The value of the color. `0.0` is black, `1.0` is full brightness.
        """
        return self.hsv[2]

    @property
    def perceived_brightness(self) -> float:
        """
        How bright the color appears to humans.

        Approximates how bright the color appears to humans. `0.0` is black,
        `1.0` is full brightness.
        """

        return math.sqrt(
            0.299 * self.red**2 + 0.587 * self.green**2 + 0.114 * self.blue**2
        )

    @property
    def hex(self) -> str:
        """
        The color as a hex string.

        The color, formatted as 8 hex digits. The first two digits are the red
        component, followed by the green component, blue and opacity.
        """

        red_hex = f"{int(round(self.red*255)):02x}"
        green_hex = f"{int(round(self.green*255)):02x}"
        blue_hex = f"{int(round(self.blue*255)):02x}"
        opacity_hex = f"{int(round(self.opacity*255)):02x}"
        return red_hex + green_hex + blue_hex + opacity_hex

    def replace(
        self,
        *,
        red: float | None = None,
        green: float | None = None,
        blue: float | None = None,
        opacity: float | None = None,
    ) -> "Color":
        """
        Return a new `Color` instance with the given values replaced.

        Return a new `Color` instance with the given values replaced. Any values
        that are not given will be copied from this color.

        ## Parameters

        red: The red component of the new color.

        green: The green component of the new color.

        blue: The blue component of the new color.

        opacity: The opacity of the new color.
        """

        return Color.from_rgb(
            red=self.red if red is None else red,
            green=self.green if green is None else green,
            blue=self.blue if blue is None else blue,
            opacity=self.opacity if opacity is None else opacity,
        )

    def _map_rgb(self, func: Callable[[float], float]) -> "Color":
        """
        Apply a function to each of the RGB values of this color, and return a
        new `Color` instance with the result. The opacity value is copied
        unchanged.
        """
        return Color.from_rgb(
            func(self.red),
            func(self.green),
            func(self.blue),
            self.opacity,
        )

    def brighter(self, amount: float) -> "Color":
        """
        Return a lighter version of this color.

        Return a new `Color` instance that is brighter than this one by the
        given amount. `0` means no change, `1` will turn the color into white.
        Values less than `0` will darken the color instead.

        How exactly the lightening/darkening happens isn't defined.

        ## Parameters

        amount: How much to lighten the color. `0` means no change, `1`
            will turn the color into white. Values less than `0` will
            darken the color instead.
        """
        # The amount may be negative. If that is the case, delegate to `darker`
        if amount <= 0:
            return self.darker(-amount)

        # HSV has an explicit value for brightness, so convert to HSV and bump
        # the value.
        #
        # Brightening by `1` is supposed to yield white, but because this
        # function starts shifting colors to white after they exceed `1.0` an
        # amount of `2` might be needed to actually get white.
        #
        # -> Apply double the amount
        hue, saturation, value = self.hsv
        value += amount * 2

        # Bumping it might put the value above 1.0. Clip it and see by how much
        # 1.0 was overshot
        value_clip = max(min(value, 1.0), 0.0)
        overshoot = value - value_clip

        # If there was an overshoot, reduce the saturation, thus pushing the
        # color towards white
        saturation = max(saturation - overshoot * 1.0, 0.0)

        return Color.from_hsv(hue, saturation, value_clip)

    def darker(self, amount: float) -> "Color":
        """
        Return a darker version of this color.

        Return a new `Color` instance that is darker than this one by the given
        amount. `0` means no change, `1` will turn the color into black. Values
        less than `0` will brighten the color instead.

        How exactly the lightening/darkening happens isn't defined.

        ## Parameters

        amount: How much to darken the color. `0` means no change, `1`
            will turn the color into black. Values less than `0` will brighten
            the color instead.
        """
        # The value may be negative. If that is the case, delegate to `brighter`
        if amount <= 0:
            return self.brighter(-amount)

        return Color._map_rgb(self, lambda x: max(x - amount, 0))

    def desaturated(self, amount: float) -> "Color":
        """
        Returns a desaturated version of this color.

        Return a copy of this color with the saturation reduced by the given
        amount. `0` means no change, `1` will turn the color into a shade of
        grey.

        ## Parameters

        amount: How much to desaturate the color. `0` means no change, `1`
            will turn the color into a shade of grey.
        """

        if amount < 0.0 or amount > 1.0:
            raise ValueError("`amount` must be between 0.0 and 1.0")

        hue, saturation, brightness = self.hsv
        saturation = saturation * (1 - amount)

        return Color.from_hsv(hue, saturation, brightness)

    def blend(self, other: Color, factor: float) -> Color:
        """
        Blend this color with another color.

        Return a new `Color` instance that is a blend of this color and the
        given `other` color. `factor` controls how much of the other color is
        used. A value of `0` will return this color, a value of `1` will return
        the other color.

        Values outside of the range `0` to `1` are allowed and will lead to the
        color being extrapolated.

        ## Parameters

        other: The other color to blend with.

        factor: How much of the other color to use. `0` will return this
            color, `1` will return the other color.
        """
        one_minus_factor = 1 - factor

        return Color.from_rgb(
            red=self.red * one_minus_factor + other.red * factor,
            green=self.green * one_minus_factor + other.green * factor,
            blue=self.blue * one_minus_factor + other.blue * factor,
            opacity=self.opacity * one_minus_factor + other.opacity * factor,
        )

    @property
    def as_plotly(self) -> str:
        """
        The color as a Plotly-compatible string.

        Plotly expects colors to be specified as strings, and this function
        returns the color formatted as such.
        """
        return f"rgba({int(round(self.red*255))}, {int(round(self.green*255))}, {int(round(self.blue*255))}, {self.opacity})"

    def _serialize(self, sess: rio.Session) -> Jsonable:
        return self.rgba

    def __repr__(self) -> str:
        return f"<Color {self.hex}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Color):
            return NotImplemented

        return self.rgba == other.rgba

    def __hash__(self) -> int:
        return hash(self.rgba)

    # Greys
    BLACK: ClassVar["Color"]
    GREY: ClassVar["Color"]
    WHITE: ClassVar["Color"]

    # Pure colors
    RED: ClassVar["Color"]
    GREEN: ClassVar["Color"]
    BLUE: ClassVar["Color"]

    # CMY
    CYAN: ClassVar["Color"]
    MAGENTA: ClassVar["Color"]
    YELLOW: ClassVar["Color"]

    # Others
    PINK: ClassVar["Color"]
    PURPLE: ClassVar["Color"]
    ORANGE: ClassVar["Color"]
    BROWN: ClassVar["Color"]

    # Special
    TRANSPARENT: ClassVar["Color"]


Color.BLACK = Color.from_rgb(0.0, 0.0, 0.0)
Color.GREY = Color.from_rgb(0.5, 0.5, 0.5)
Color.WHITE = Color.from_rgb(1.0, 1.0, 1.0)

Color.RED = Color.from_rgb(1.0, 0.0, 0.0)
Color.GREEN = Color.from_rgb(0.0, 1.0, 0.0)
Color.BLUE = Color.from_rgb(0.0, 0.0, 1.0)

Color.CYAN = Color.from_rgb(0.0, 1.0, 1.0)
Color.MAGENTA = Color.from_rgb(1.0, 0.0, 1.0)
Color.YELLOW = Color.from_rgb(1.0, 1.0, 0.0)

Color.PINK = Color.from_rgb(1.0, 0.0, 1.0)
Color.PURPLE = Color.from_rgb(0.5, 0.0, 0.5)
Color.ORANGE = Color.from_rgb(1.0, 0.5, 0.0)

Color.TRANSPARENT = Color.from_rgb(0.0, 0.0, 0.0, 0.0)


# Like color, but also allows referencing theme colors
ColorSet: TypeAlias = (
    Color
    | Literal[
        "background",
        "neutral",
        "hud",
        "primary",
        "secondary",
        "success",
        "warning",
        "danger",
        "keep",
    ]
)


# Cache so the session can quickly determine whether a type annotation is
# `ColorSet`
_color_set_args = set(get_args(ColorSet))



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Any, final

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "ColorPicker",
    "ColorChangeEvent",
]


@final
@dataclass
class ColorChangeEvent:
    """
    The argument for the `on_change` handler of a `ColorPicker`.

    ## Attributes

    `color`: The new `color` of the `ColorPicker`.
    """

    color: rio.Color


@final
class ColorPicker(FundamentalComponent):
    """
    Allows the user to pick a RGB(A) color.

    `ColorPicker` is a component that allows the user to input a color. It
    displays a combination of colorful areas and sliders that the user can
    interact with to pick a color, and optionally an opacity slider to pick
    opacity.


    ## Attributes

    `color`: The color that the user has picked.

    `pick_opacity`: Whether to allow the user to pick opacity. If `False`,
        the opacity slider will be hidden and the color value will be forced
        to be fully opaque.

    `on_change`: This event is triggered whenever the user changes the color.


    ## Examples

    This small example will print a message whenever the user changes the
    selected color:

    ```python
    def print_selected_color(event: rio.ColorChangeEvent):
        print('You selected the color', event.color)

    rio.ColorPicker(
        rio.Color.from_hex("#ff0000"),
        on_change=print_selected_color,
    )
    ```

    This one utilizes a attribute binding to use the selected color as the fill for
    an `Icon`:

    ```python
    class ButtonColorizer(rio.Component):
        color: rio.Color = rio.Color.BLUE

        def build(self) -> rio.Component:
            return rio.Row(
                rio.ColorPicker(
                    color=self.bind().color,
                ),
                rio.Icon(
                    'material/star',
                    fill=self.color,
                )
            )
    ```
    """

    color: rio.Color
    _: KW_ONLY
    pick_opacity: bool = False
    on_change: rio.EventHandler[ColorChangeEvent] = None

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        color = rio.Color.from_rgb(*msg["color"])

        # Update the color
        self._apply_delta_state_from_frontend({"color": color})

        # Trigger the change event
        await self.call_event_handler(
            self.on_change,
            ColorChangeEvent(color),
        )

        # Refresh the session
        await self.session._refresh()

    def _custom_serialize(self) -> JsonDoc:
        return {
            "color": self.color.rgba,
        }


ColorPicker._unique_id = "ColorPicker-builtin"



import pytest
from utils import create_mockapp

import rio


async def test_session_attachments():
    async with create_mockapp() as app:
        session = app.session

        list1 = ["foo", "bar"]
        list2 = []

        session.attach(list1)
        assert session[list] is list1

        session.attach(list2)
        assert session[list] is list2


async def test_access_nonexistent_session_attachment():
    async with create_mockapp() as app:
        session = app.session

        with pytest.raises(KeyError):
            session[list]


async def test_default_attachments():
    class Settings(rio.UserSettings):
        foo: int

    dict_attachment = {"foo": "bar"}
    settings_attachment = Settings(3)

    async with create_mockapp(
        default_attachments=[dict_attachment, settings_attachment]
    ) as app:
        session = app.session

        # Default attachments shouldn't be copied, unless they're UserSettings
        assert session[dict] is dict_attachment

        assert session[Settings] == settings_attachment
        assert session[Settings] is not settings_attachment



from uniserde import JsonDoc

from .fundamental_component import FundamentalComponent

__all__ = ["BuildFailed"]


class BuildFailed(FundamentalComponent):
    """
    Used as a placeholder in case a component's `build` function throws an
    error.


    ## Metadata

    public: False
    """

    error_summary: str
    error_details: str

    # Debug details can contain sensitive information. Sending these to the
    # client is fine during development, but mustn't happen in production.
    def _custom_serialize(self) -> JsonDoc:
        if self.session._app_server.debug_mode:
            return {
                "_rio_internal_": True,
            }
        else:
            return {
                "error_summary": "This component has crashed",
                "error_details": "",
                "_rio_internal_": True,
            }


BuildFailed._unique_id = "BuildFailed-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from .component import Component
from .fundamental_component import FundamentalComponent
from .progress_circle import ProgressCircle

__all__ = [
    "Button",
    "IconButton",
]


CHILD_MARGIN_X = 1.0
CHILD_MARGIN_Y = 0.3
INITIALLY_DISABLED_FOR = 0.25


@final
class Button(Component):
    """
    A clickable button.

    The `Button` component allows the user to trigger an action by clicking on
    it. You can use it to trigger a function call, navigate to a different page,
    or perform other actions.


    ## Attributes

    `content`: The text or child component to display inside of the button.

    `icon`: The name of an icon to display on the button, in the form
        "set/name:variant". See the `Icon` component for details of how
        icons work in Rio.

    `shape`: The shape of the button. This can be one of:

        - `"pill"`: A rectangle where the left and right sides are completely
            round.
        - `"rounded"`: A rectangle with rounded corners.
        - `"rectangle"`: A rectangle with sharp corners.

    `style`: Controls the button's appearance. This can be one of:

        - `"major"`: A highly visible button with bold visuals.
        - `"minor"`: A less visible button that blends into the background.
        - `"plain"`: A button with no background or border. Use this to make the
            button look like a link.

    `color`: The color scheme to use for the button.

    `is_sensitive`: Whether the button should respond to user input.

    `is_loading`: Whether the button should display a loading indicator. Use
        this to indicate to the user that an action is currently running.

    `initially_disabled_for`: The number of seconds the button should be
        disabled for after it is first rendered. This is useful to prevent
        the user from accidentally pressing a button that suddenly appeared.

    `on_press`: Triggered when the user clicks on the button.


    ## Examples

    This code creates a button with the caption "Click me!":

    ```python
    rio.Button(
        "Click me!",
        on_press=lambda: print("Button pressed!"),
    )
    ```

    You can make it a little fancier by adding an icon:

    ```python
    rio.Button(
        "Click me!",
        icon="material/mouse",
        on_press=lambda: print("Button pressed!"),
    )
    ```

    You can even put other components inside of the button. Here's a button with
    a progress bar that slowly fills up as you click it:

    ```python
    class ProgressButton(rio.Component):
        clicks: int = 0

        def _on_button_press(self) -> None:
            self.clicks += 1

        def build(self) -> rio.Component:
            return rio.Button(
                rio.Column(
                    rio.Text("Click repeatedly to fill up the progress bar"),
                    rio.ProgressBar(self.clicks/10, width=15, height=1),
                ),
                on_press=self._on_button_press,
            )
    ```
    """

    content: str | rio.Component = ""
    _: KW_ONLY
    icon: str | None = None
    shape: Literal["pill", "rounded", "rectangle"] = "pill"
    style: Literal["major", "minor", "plain"] = "major"
    color: rio.ColorSet = "keep"
    is_sensitive: bool = True
    is_loading: bool = False
    initially_disabled_for: float = INITIALLY_DISABLED_FOR
    on_press: rio.EventHandler[[]] = None

    def build(self) -> rio.Component:
        # Prepare the child
        if self.is_loading:
            if self.color in ("keep", "secondary"):
                progress_color = "primary"
            else:
                progress_color = "secondary"

            child = ProgressCircle(
                size=1.5,
                align_x=0.5,
                margin_x=CHILD_MARGIN_Y,
                margin_y=CHILD_MARGIN_Y,
                color=progress_color,
            )
        elif isinstance(self.content, Component):
            child = rio.Container(
                self.content,
                margin_x=CHILD_MARGIN_Y,
                margin_y=CHILD_MARGIN_Y,
                align_x=0.5,
            )
        else:
            children = []
            text = self.content.strip()
            n_children = (self.icon is not None) + bool(text)

            if self.icon is not None:
                children.append(
                    rio.Icon(
                        self.icon,
                        width=1.4,
                        height=1.4,
                        margin_x=CHILD_MARGIN_X if n_children == 1 else None,
                        margin_y=CHILD_MARGIN_Y if n_children == 1 else None,
                    )
                )

            if text:
                children.append(
                    rio.Text(
                        text,
                        margin_x=CHILD_MARGIN_X if n_children == 1 else None,
                        margin_y=CHILD_MARGIN_Y if n_children == 1 else None,
                        style=(
                            rio.TextStyle(font_weight="bold")
                            if self.style == "major"
                            else "text"
                        ),
                    )
                )

            if len(children) == 1:
                child = children[0]
            else:
                child = rio.Row(
                    *children,
                    spacing=0.6,
                    margin_x=CHILD_MARGIN_X,
                    margin_y=CHILD_MARGIN_Y,
                    align_x=0.5,
                )

        # Delegate to a HTML Component
        has_content = not isinstance(self.content, str) or self.content

        return _ButtonInternal(
            on_press=self.on_press,
            content=child,
            shape=self.shape,
            style=self.style,
            color=self.color,
            is_sensitive=self.is_sensitive,
            is_loading=self.is_loading,
            initially_disabled_for=self.initially_disabled_for,
            width=8 if has_content else 4,
            height=2.2,
        )

    def __str__(self) -> str:
        if isinstance(self.content, str):
            content = f"text:{self.content!r}"
        else:
            content = f"content:{self.content._id}"

        return f"<Button id:{self._id} {content}>"


@final
class IconButton(Component):
    """
    # IconButton

    A round, clickable button with an icon.

    The `IconButton` component allows the user to trigger an action by clicking
    on it. You can use it to trigger a function call, navigate to a different
    page, or perform other actions.

    It is similar to the `Button` component, but it is specifically designed to
    display an icon, and it has a round shape.


    ## Attributes

    `icon`: The name of an icon to display on the button, in the form
            `"set/name:variant"`. See the `Icon` component for details of how
            icons work in Rio.

    `style`: Controls the button's appearance. This can be one of:

        - `major`: A highly visible button with bold visuals.
        - `minor`: A less visible button that blends into the background.
        - `plain`: A button with no background or border. Use this to make
                        the button look like a link.

    `color`: The color scheme to use for the button.

    `is_sensitive`: Whether the button should respond to user input.

    `initially_disabled_for`: The number of seconds the button should be
            disabled for after it is first rendered. This is useful to prevent
            the user from accidentally triggering an action when the page is
            first loaded.

    `size`: The size of the button. This is the diameter of the button in
            font-size units.

    `on_press`: Triggered when the user clicks on the button.


    ## Examples

    This minimal example will simply display a `IconButton` with a castle icon:

    ```python
    rio.IconButton(icon="material/castle")
    ```

    `IconButton`s are commonly used to trigger actions. You can easily achieve this by
    adding a function call to `on_press`:

    ```python
    class MyComponent(rio.Component):
        def on_press_button(self) -> None:
            print("Icon button pressed!")

        def build(self) -> rio.Component:
            return rio.IconButton(
                icon="material/castle",
                on_press=self.on_press_button,
            )
    ```

    `IconButton`s are commonly used to trigger actions. You can easily achieve this by
    adding a function call to `on_press`. You can use a function call to update
    the banner text signaling that the button was pressed:

    ```python
    class MyComponent(rio.Component):
        banner_text: str = ""

        def on_press_button(self) -> None:
            self.banner_text = "Icon button pressed!"

        def build(self) -> rio.Component:
            return rio.Column(
                rio.Banner(
                    text=self.banner_text,
                    style="material/info",
                ),
                rio.IconButton(
                    icon="material/castle",
                    on_press=self.on_press_button,
                ),
                spacing=1,
            )
    ```
    """

    icon: str
    _: KW_ONLY
    style: Literal["major", "minor", "plain"]
    color: rio.ColorSet
    is_sensitive: bool
    initially_disabled_for: float = INITIALLY_DISABLED_FOR
    size: float
    on_press: rio.EventHandler[[]]

    def __init__(
        self,
        icon: str,
        *,
        style: Literal["major", "minor", "plain"] = "major",
        color: rio.ColorSet = "keep",
        is_sensitive: bool = True,
        on_press: rio.EventHandler[[]] = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        size: float = 3.7,
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width="natural",
            height="natural",
            align_x=align_x,
            align_y=align_y,
        )

        self.icon = icon
        self.size = size
        self.style = style
        self.color = color
        self.is_sensitive = is_sensitive
        self.on_press = on_press
        self.initially_disabled_for = INITIALLY_DISABLED_FOR

    def build(self) -> rio.Component:
        return _ButtonInternal(
            on_press=self.on_press,
            content=rio.Icon(
                self.icon,
                height=self.size * 0.65,
                width=self.size * 0.65,
                align_x=0.5,
                align_y=0.5,
            ),
            shape="circle",
            style=self.style,
            color=self.color,
            is_sensitive=self.is_sensitive,
            is_loading=False,
            width=self.size,
            height=self.size,
            initially_disabled_for=self.initially_disabled_for,
            # Make sure the button has a square aspect ratio
            align_x=0.5,
            align_y=0.5,
        )

    def _get_debug_details(self) -> dict[str, Any]:
        result = super()._get_debug_details()

        # `width` & `height` are replaced with `size`
        del result["width"]
        del result["height"]

        return result


class _ButtonInternal(FundamentalComponent):
    _: KW_ONLY
    on_press: rio.EventHandler[[]]
    content: rio.Component
    shape: Literal["pill", "rounded", "rectangle", "circle"]
    style: Literal["major", "minor", "plain"]
    color: rio.ColorSet
    is_sensitive: bool
    is_loading: bool
    initially_disabled_for: float

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg
        assert msg["type"] == "press", msg

        msg_type: str = msg["type"]
        assert isinstance(msg_type, str), msg_type

        # Is the button sensitive?
        if not self.is_sensitive or self.is_loading:
            return

        # Trigger the press event
        await self.call_event_handler(self.on_press)

        # Refresh the session
        await self.session._refresh()


_ButtonInternal._unique_id = "Button-builtin"



"""
Contains documentation related tasks specific to the Rio project.
"""

import dataclasses
import functools
import inspect
from typing import *  # type: ignore

import imy.docstrings
import unicall

import rio

__all__ = [
    "get_docs_for",
    "get_documentation_fragment",
    "build_documentation_url",
    "find_documented_objects",
    "postprocess_function_docs",
    "postprocess_class_docs",
    "postprocess_component_docs",
]


def _make_docs_for_rio_url():
    docs = imy.docstrings.ClassDocs.from_class(rio.URL)
    docs.attributes.clear()
    docs.functions.clear()
    docs.summary = "Alias for `yarl.URL`."
    docs.details = """
Since URLs are a commonly used data type, `rio` re-exports `yarl.URL` as
`rio.URL` for convenience. See the `yarl` documentation
[here](https://yarl.aio-libs.org/en/stable/api/#yarl.URL) for details about
this class.
""".strip()

    return docs


DOCS_OVERRIDES: dict[
    Any, imy.docstrings.ClassDocs | imy.docstrings.FunctionDocs
] = {
    rio.URL: _make_docs_for_rio_url(),
}


@functools.lru_cache(maxsize=None)
def get_docs_for(
    obj: Callable | Type,
) -> imy.docstrings.ClassDocs | imy.docstrings.FunctionDocs:
    """
    Parse the docs for a component and return them. The results are cached, so
    this function is fast.
    """

    try:
        return DOCS_OVERRIDES[obj]
    except KeyError:
        pass

    # Components and other classes
    if inspect.isclass(obj):
        docs = imy.docstrings.ClassDocs.from_class(obj)

        # Apply rio-specific post-processing
        if issubclass(obj, rio.Component):
            postprocess_component_docs(docs)
        else:
            postprocess_class_docs(docs)

        return docs

    # Functions
    if callable(obj):
        docs = imy.docstrings.FunctionDocs.from_function(obj)

        # Apply rio-specific post-processing
        postprocess_function_docs(docs)

        return docs

    # No clue
    raise ValueError(f"Cannot get docs for object `{obj}`")


def get_documentation_fragment(object_name: str) -> str:
    """
    Returns the fragment part of the URL corresponding to the documentation for
    the given Rio object.
    """
    return object_name.lower()


def build_documentation_url(
    object_name: str,
    *,
    relative: bool = False,
) -> rio.URL:
    """
    Returns the URL to the documentation for the given Rio object. This doesn't
    perform any checks on whether the object actually exists and has
    documentation. It relies solely on the passed values.
    """

    # Build the relative URL
    result_string = f"/docs/api/{get_documentation_fragment(object_name)}"

    # Make it absolute, if requested
    if not relative:
        result_string = "https://rio.dev" + result_string

    # Done
    return rio.URL(result_string)


def _find_possibly_public_objects() -> Iterable[Type | Callable]:
    """
    Finds all objects in rio that might be public. This uses heuristics to
    filter out many internal objects, but it's not perfect.
    """
    # Hardcoded items
    yield rio.App
    yield rio.AssetError
    yield rio.Color
    yield rio.ColorChangeEvent
    yield rio.CursorStyle
    yield rio.DrawerOpenOrCloseEvent
    yield rio.DropdownChangeEvent
    yield rio.escape_markdown_code
    yield rio.escape_markdown
    yield rio.FileInfo
    yield rio.Font
    yield rio.KeyDownEvent
    yield rio.KeyPressEvent
    yield rio.KeyUpEvent
    yield rio.MouseDownEvent
    yield rio.MouseEnterEvent
    yield rio.MouseLeaveEvent
    yield rio.MouseMoveEvent
    yield rio.MouseUpEvent
    yield rio.NavigationFailed
    yield rio.NumberInputChangeEvent
    yield rio.NumberInputConfirmEvent
    yield rio.Page
    yield rio.PopupOpenOrCloseEvent
    yield rio.RevealerChangeEvent
    yield rio.Session
    yield rio.TextInputChangeEvent
    yield rio.TextInputConfirmEvent
    yield rio.TextStyle
    yield rio.Theme
    yield rio.UserSettings

    # Yield classes that also need their children documented
    to_do = [
        rio.Fill,
        rio.Component,
    ]

    while to_do:
        cur = to_do.pop()

        # Queue the children
        to_do.extend(cur.__subclasses__())

        # Hardcoded items that DON'T need documentation
        if cur.__name__ in (
            "ClassContainer",
            "FundamentalComponent",
            "RootContainer",
        ):
            continue

        # Skip anything not in the `rio` module
        if not cur.__module__.startswith("rio."):
            continue

        # Internal
        if cur.__name__.startswith("_"):
            continue

        yield cur


def find_documented_objects(
    postprocess: bool,
) -> Iterable[
    tuple[Type, imy.docstrings.ClassDocs]
    | tuple[Callable, imy.docstrings.FunctionDocs],
]:
    """
    Find all public classes and functions in `Rio` along with their
    documentation.

    ## Parameters

    `postprocess`: Whether to apply rio-specific postprocessing to the
    documentation.
    """
    # Use heuristics to find all objects which should likely be public
    for obj in _find_possibly_public_objects():
        # Parse the object's docs
        if inspect.isclass(obj):
            docs = imy.docstrings.ClassDocs.from_class(obj)

            if postprocess:
                if issubclass(obj, rio.Component):
                    postprocess_component_docs(docs)
                else:
                    postprocess_class_docs(docs)

        elif callable(obj):
            docs = imy.docstrings.FunctionDocs.from_function(obj)

            if postprocess:
                postprocess_function_docs(docs)

        else:
            raise ValueError(f"Unexpected object type: {obj}")

        # Make the final determination whether this object is public
        if not docs.metadata.public:
            continue

        # This object is public
        yield obj, docs  # type: ignore


def postprocess_function_docs(docs: imy.docstrings.FunctionDocs) -> None:
    # Nothing here yet. This function exists for compatibility. If we ever
    # decide to add function-specific post-processing, we can do it here without
    # having to change all scripts.
    pass


def postprocess_class_docs(docs: imy.docstrings.ClassDocs) -> None:
    """
    Perform Rio specific post-processing on the component, such as stripping out
    internal attributes and functions.
    """

    # Strip out anything `Session` inherits from `unicall`
    if docs.name == "Session":
        to_remove = set(dir(unicall.Unicall)).difference(vars(rio.Session))
        docs.functions = [
            func for func in docs.functions if func.name not in to_remove
        ]

    # Strip default docstrings created by dataclasses
    if docs.summary is not None and docs.summary.startswith(f"{docs.name}("):
        docs.summary = None
        docs.details = None

    # Strip internal attributes
    index = 0
    while index < len(docs.attributes):
        attr = docs.attributes[index]

        # Decide whether to keep it
        keep = True

        keep = keep and not attr.name.startswith("_")
        keep = keep and attr.type is not dataclasses.KW_ONLY

        # Strip it out, if necessary
        if keep:
            index += 1
        else:
            del docs.attributes[index]

    # Skip internal functions
    index = 0
    while index < len(docs.functions):
        func = docs.functions[index]

        # Decide whether to keep it

        # Internal methods start with an underscore
        keep = not func.name.startswith("_")

        # Some methods in components are meant to be used by the user, but only
        # when they're the one creating the component. For example, the `build`
        # method is only intended to be used by the component itself, and
        # documenting it would be pointless at best, and confusing at worst.
        is_inherited_protected_method = (
            docs.name != "Component"
            and func.name
            in (
                "build",
                "call_event_handler",
                "force_refresh",
            )
        )
        if is_inherited_protected_method:
            keep = False

        # Strip lambdas
        if func.name == "<lambda>":
            keep = False

        # Make sure to keep the constructor
        if func.name == "__init__":
            keep = True

        # Some classes are not meant to be constructed by the user. Strip their
        # constructor.
        if (
            docs.name
            in (
                "FileInfo",
                "Session",
            )
            and func.name == "__init__"
        ):
            keep = False

        # Check if it's explicitly excluded
        if not docs.metadata.public:
            keep = False

        # Strip it out, if necessary
        if keep:
            index += 1
        else:
            del docs.functions[index]

    # Additional per-function post-processing
    for func_docs in docs.functions:
        # Strip the ridiculous default docstring created by dataclasses
        #
        # FIXME: Not working for some reason
        if (
            func_docs.name == "__init__"
            and func_docs.summary
            == "Initialize self. See help(type(self)) for accurate signature."
        ):
            func_docs.summary = None
            func_docs.details = None

        # Inject a short description for `__init__` if there is none.
        if func_docs.name == "__init__" and func_docs.summary is None:
            func_docs.summary = f"Creates a new `{docs.name}` instance."


def postprocess_component_docs(docs: imy.docstrings.ClassDocs) -> None:
    # Apply the standard class post-processing
    postprocess_class_docs(docs)

    # Remove the `bind()` method inherited from `rio.Component`, because
    # that method is only useful in custom components
    for i, func in enumerate(docs.functions):
        if func.name == "bind":
            del docs.functions[i]
            break



import rio

# <additional-imports>
from .. import data_models

# </additional-imports>


# <component>
class ItemEditor(rio.Component):
    """
    A component for editing a menu item.

    Returns a card component containing the menu item editor. The editor
    contains fields for name, description, price, and category of the menu item.
    The component also contains buttons for saving or canceling the changes.

    The on_change methods are event handlers for the respective input fields.
    These methods update the attributes of the currently selected menu item.

    ## Attributes

    `currently_selected_menu_item`: The currently selected menu item.

    `new_entry`: A boolean flag indicating if the menu item is a new entry.

    `on_cancel_event`: An event handler for the cancel button.

    `on_save_event`: An event handler for the save button.
    """

    currently_selected_menu_item: data_models.MenuItems
    new_entry: bool = False
    on_cancel_event: rio.EventHandler[[]] = None
    on_save_event: rio.EventHandler[[]] = None

    def on_change_name(self, ev: rio.TextInputChangeEvent) -> None:
        """
        Changes the name of the currently selected menu item. And updates the
        name attribute of our data model.

        ## Parameters

        `ev`: The event object that contains the new name.
        """
        self.currently_selected_menu_item.name = ev.text

    def on_change_description(self, ev: rio.TextInputChangeEvent) -> None:
        """
        Changes the description of the currently selected menu item. And updates
        the description attribute of our data model.

        ## Parameters

        `ev`: The event object that contains the new description.
        """
        self.currently_selected_menu_item.description = ev.text

    def on_change_price(self, ev: rio.NumberInputChangeEvent) -> None:
        """
        Changes the price of the currently selected menu item. And updates the
        price attribute of our data model.

        ## Parameters
        `ev`: The event object that contains the new price.
        """
        self.currently_selected_menu_item.price = ev.value

    def on_change_category(self, ev: rio.DropdownChangeEvent) -> None:
        """
        Changes the category of the currently selected menu item. And updates
        the category attribute of our data model.

        ## Parameters
        `ev`: The event object that contains the new category.
        """
        self.currently_selected_menu_item.category = ev.value

    def build(self) -> rio.Component:
        """
        Builds the menu item editor component.

        See the approx. layout below:

        ```
        ╔══════════════════════ Card ══════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━ Text ━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Edit Menu Item | Add New Menu Item           ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━ TextInput ━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Name                                         ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━ TextInput ━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Description                                  ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━ NumberInput ━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Price                                        ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━ Dropdown ━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Category                                     ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━━━━ Row ━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃   ┏━━━ Button ━━━┓        ┏━━━ Button ━━━┓   ┃ ║
        ║ ┃   ┃ Save         ┃        ┃ Cancel       ┃   ┃ ║
        ║ ┃   ┗━━━━━━━━━━━━━━┛        ┗━━━━━━━━━━━━━━┛   ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚══════════════════════════════════════════════════╝
        ```
        """

        if self.new_entry is False:
            text = "Edit Menu Item"
        else:
            text = "Add New Menu Item"

        return rio.Card(
            rio.Column(
                rio.Text(
                    text=text,
                    style="heading2",
                    margin_bottom=1,
                ),
                rio.TextInput(
                    self.currently_selected_menu_item.name,
                    label="Name",
                    on_change=self.on_change_name,
                ),
                rio.TextInput(
                    self.currently_selected_menu_item.description,
                    label="Description",
                    on_change=self.on_change_description,
                ),
                rio.NumberInput(
                    self.currently_selected_menu_item.price,
                    label="Price",
                    suffix_text="$",
                    on_change=self.on_change_price,
                ),
                rio.Dropdown(
                    options=[
                        "Burgers",
                        "Desserts",
                        "Drinks",
                        "Salads",
                        "Sides",
                    ],
                    label="Category",
                    selected_value=self.currently_selected_menu_item.category,
                    on_change=self.on_change_category,
                ),
                rio.Row(
                    rio.Button("Save", on_press=self.on_save_event),
                    rio.Button("Cancel", on_press=self.on_cancel_event),
                    spacing=1,
                    align_x=1,
                ),
                spacing=1,
                align_y=0,
                margin=2,
            ),
        )


# </component>



from __future__ import annotations

import logging
from collections.abc import Callable, Iterable
from dataclasses import KW_ONLY, dataclass, field
from typing import final

import rio

from . import common
from .errors import NavigationFailed

__all__ = ["Page"]


@final
@dataclass(frozen=True)
class Page:
    """
    A routable page in a Rio app.

    Rio apps can consist of many pages. You might have a welcome page, a
    settings page, a login, and so on. `Page` components contain all information
    needed to display those pages, as well as to navigate between them.

    This is not just specific to websites. Apps might, for example, have
    a settings page, a profile page, a help page, and so on.

    Pages are passed directly to the app during construction, like so:

    ```python
    import rio

    app = rio.App(
        build=lambda: rio.Column(
            rio.Text("Welcome to my app!"),
            rio.PageView(height="grow"),
        ),
        pages=[
            rio.Page(
                name="Home",
                page_url="",
                build=lambda: rio.Text("This is the home page"),
            ),
            rio.Page(
                name="Subpage",
                page_url="subpage",
                build=lambda: rio.Text("This is a subpage"),
            ),
        ],
    )

    app.run_in_browser()
    ```

    This will display "This is the home page" when navigating to the root URL,
    but "This is a subpage" when navigating to "/subpage". Note that on both
    pages the text "Welcome to my page!" is displayed above the page content.
    That's because it's not part of the `PageView`.

    TODO: Link to the routing/multipage how-to page

    ## Attributes

    `name`: A human-readable name for the page. While the page itself doesn't
        use this value directly, it serves as important information for
        debugging, as well as other components such as navigation bars.

    `page_url`: The URL segment at which this page should be displayed. For
        example, if this is "subpage", then the page will be displayed at
        "https://yourapp.com/subpage". If this is "", then the page will be
        displayed at the root URL.

    `build`: A callback that is called when this page is displayed. It should
        return a Rio component.

    `icon`: The name of an icon to associate with the page. While the page
        itself doesn't use this value directly, it serves as additional
        information for other components such as navigation bars.

    `children`: A list of child pages. These pages will be displayed when
        navigating to a sub-URL of this page. For example, if this page's
        `page_url` is "page1", and it has a child page with `page_url` "page2",
        then the child page will be displayed at
        "https://yourapp.com/page1/page2".

    `guard`: A callback that is called before this page is displayed. It
        can prevent users from accessing pages which they are not allowed to
        see. For example, you may want to redirect users to your login page
        if they are trying to access their profile page without being
        logged in.

        The callback should return `None` if the user is allowed to access
        the page, or a string or `rio.URL` if the user should be redirected
        to a different page.
    """

    name: str
    page_url: str
    build: Callable[[], rio.Component]
    _: KW_ONLY
    icon: str = "rio/logo:color"
    children: list[Page] = field(default_factory=list)
    guard: (
        Callable[[rio.Session, tuple[rio.Page, ...]], None | rio.URL | str]
        | None
    ) = None

    def __post_init__(self) -> None:
        # URLs are case insensitive. An easy way to enforce this, and also
        # prevent casing issues in the user code is to make sure the page's URL
        # fragment is lowercase.
        if self.page_url != self.page_url.lower():
            raise ValueError(
                f"Page URLs have to be lowercase, but `{self.page_url}` is not"
            )


def _get_active_page_instances(
    available_pages: Iterable[rio.Page],
    remaining_segments: tuple[str, ...],
) -> list[rio.Page]:
    """
    Given a list of available pages, and a URL, return the list of pages that
    would be active if navigating to that URL.
    """
    # Get the page responsible for this segment
    try:
        page_segment = remaining_segments[0]
    except IndexError:
        page_segment = ""

    for page in available_pages:
        if page.page_url == page_segment:
            break
    else:
        return []

    # Recurse into the children
    sub_pages = _get_active_page_instances(
        page.children,
        remaining_segments[1:],
    )
    return [page] + sub_pages


def check_page_guards(
    sess: rio.Session,
    target_url_absolute: rio.URL,
) -> tuple[tuple[Page, ...], rio.URL]:
    """
    Check whether navigation to the given target URL is possible.

    This finds the pages that would be activated by this navigation and runs
    their guards. If the guards effect a redirect, it instead attempts to
    navigate to the redirect target, and so on.

    Raises `NavigationFailed` if navigation to the target URL is not possible
    because of an error, such as an exception in a guard.

    If the URL points to a page which doesn't exist that is not considered an
    error. The result will still be valid. That is because navigation is
    possible, it's just that some PageViews will display a 404 page.

    This function does not perform any actual navigation. It simply checks
    whether navigation to the target page is possible.
    """

    assert target_url_absolute.is_absolute(), target_url_absolute

    # Is any guard opposed to this page?
    initial_target_url = target_url_absolute
    visited_redirects = {target_url_absolute}
    past_redirects = [target_url_absolute]

    while True:
        # TODO: What if the URL is not a child of the base URL? i.e. redirecting
        #   to a completely different site
        target_url_relative = common.make_url_relative(
            sess._base_url, target_url_absolute
        )

        # Find all pages which would by activated by this navigation
        active_page_instances = tuple(
            _get_active_page_instances(
                sess.app.pages, target_url_relative.parts
            )
        )

        # Check the guards for each activated page
        redirect = None
        for page in active_page_instances:
            if page.guard is None:
                continue

            try:
                redirect = page.guard(sess, active_page_instances)
            except Exception as err:
                message = f"Rejecting navigation to `{initial_target_url}` because of an error in a page guard of `{page.page_url}`: {err}"
                logging.exception(message)
                raise NavigationFailed(message)

            if redirect is not None:
                break

        # All guards are satisfied - done
        if redirect is None:
            return active_page_instances, target_url_absolute

        # A guard wants to redirect to a different page
        if isinstance(redirect, str):
            redirect = rio.URL(redirect)

        redirect = sess._active_page_url.join(redirect)

        assert redirect.is_absolute(), redirect

        # Detect infinite loops and break them
        if redirect in visited_redirects:
            page_strings = [
                str(page_url) for page_url in past_redirects + [redirect]
            ]
            page_strings_list = "\n -> ".join(page_strings)

            message = f"Rejecting navigation to `{initial_target_url}` because page guards have created an infinite loop:\n\n    {page_strings_list}"
            logging.warning(message)
            raise NavigationFailed(message)

        # Remember that this page has been visited before
        visited_redirects.add(redirect)
        past_redirects.append(redirect)

        # Rinse and repeat
        target_url_absolute = redirect



from __future__ import annotations

import pathlib
from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import common
from .fills import FillLike
from .self_serializing import SelfSerializing

__all__ = [
    "Font",
    "TextStyle",
]


class UnsetType:
    pass


UNSET = UnsetType()


@dataclass(frozen=True)
class Font(SelfSerializing):
    """
    A custom font face.

    The `Font` class lets you create custom fonts for use in your rio app. To
    instantiate a `Font`, you must pass it at least one font file. (As far as
    rio is concerned, the file format is irrelevant - all that matters is that
    a browser can display it.)


    ## Attributes

    `regular`: The regular (i.e. not bold, not italic) font file.

    `bold`: The bold font file.

    `italic`: The italic font file.

    `bold_italic`: The bold and italic font file.
    """

    regular: pathlib.Path | bytes
    bold: pathlib.Path | bytes | None = None
    italic: pathlib.Path | bytes | None = None
    bold_italic: pathlib.Path | bytes | None = None

    def _serialize(self, sess: rio.Session) -> str:
        return sess._register_font(self)

    # Predefined fonts
    ROBOTO: ClassVar[Font]
    ROBOTO_MONO: ClassVar[Font]


Font.ROBOTO = Font(
    regular=common.HOSTED_ASSETS_DIR / "fonts/Roboto/Roboto-Regular.ttf",
    bold=common.HOSTED_ASSETS_DIR / "fonts/Roboto/Roboto-Bold.ttf",
    italic=common.HOSTED_ASSETS_DIR / "fonts/Roboto/Roboto-Italic.ttf",
    bold_italic=common.HOSTED_ASSETS_DIR / "fonts/Roboto/Roboto-BoldItalic.ttf",
)

Font.ROBOTO_MONO = Font(
    regular=common.HOSTED_ASSETS_DIR
    / "fonts/Roboto Mono/RobotoMono-Regular.ttf",
    bold=common.HOSTED_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Bold.ttf",
    italic=common.HOSTED_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Italic.ttf",
    bold_italic=common.HOSTED_ASSETS_DIR
    / "fonts/Roboto Mono/RobotoMono-BoldItalic.ttf",
)


@dataclass(frozen=True)
class TextStyle(SelfSerializing):
    _: KW_ONLY
    font: Font | None = None
    fill: FillLike | None = None
    font_size: float = 1.0
    italic: bool = False
    font_weight: Literal["normal", "bold"] = "normal"
    underlined: bool = False
    all_caps: bool = False

    def replace(
        self,
        *,
        font: Font | None = None,
        fill: FillLike | None | UnsetType = UNSET,
        font_size: float | None = None,
        italic: bool | None = None,
        font_weight: Literal["normal", "bold"] | None = None,
        underlined: bool | None = None,
        all_caps: bool | None = None,
    ) -> TextStyle:
        return type(self)(
            font=self.font if font is None else font,
            fill=self.fill if isinstance(fill, UnsetType) else fill,
            font_size=self.font_size if font_size is None else font_size,
            italic=self.italic if italic is None else italic,
            font_weight=self.font_weight
            if font_weight is None
            else font_weight,
            underlined=self.underlined if underlined is None else underlined,
            all_caps=self.all_caps if all_caps is None else all_caps,
        )

    def _serialize(self, sess: rio.Session) -> JsonDoc:
        return {
            "fontName": None
            if self.font is None
            else self.font._serialize(sess),
            "fill": None if self.fill is None else self.fill._serialize(sess),
            "fontSize": self.font_size,
            "italic": self.italic,
            "fontWeight": self.font_weight,
            "underlined": self.underlined,
            "allCaps": self.all_caps,
        }



import asyncio

from utils import create_mockapp

import rio


class ChildToggler(rio.Component):
    child: rio.Component
    switch: bool = True

    def toggle(self) -> None:
        self.switch = not self.switch

    def build(self) -> rio.Component:
        if self.switch:
            return rio.Spacer()
        else:
            return self.child


async def test_mounted():
    mounted = unmounted = False

    class DemoComponent(rio.Component):
        @rio.event.on_mount
        def _on_mount(self):
            nonlocal mounted
            mounted = True

        @rio.event.on_unmount
        def _on_unmount(self):
            nonlocal unmounted
            unmounted = True

        def build(self) -> rio.Component:
            return rio.Text("hi")

    def build():
        return ChildToggler(DemoComponent())

    async with create_mockapp(build) as app:
        root = app.get_component(ChildToggler)
        assert not mounted
        assert not unmounted

        root.toggle()
        await app.refresh()
        assert mounted
        assert not unmounted

        root.toggle()
        await app.refresh()
        assert unmounted


async def test_refresh_after_synchronous_mount_handler():
    class DemoComponent(rio.Component):
        mounted: bool = False

        @rio.event.on_mount
        def on_mount(self):
            self.mounted = True

        def build(self) -> rio.Component:
            return rio.Switch(self.mounted)

    async with create_mockapp(DemoComponent) as app:
        demo_component = app.get_component(DemoComponent)
        switch = app.get_component(rio.Switch)

        # TODO: I don't know how we can wait for the refresh, so I'll just use a
        # sleep()
        await asyncio.sleep(0.5)
        assert demo_component.mounted

        last_component_state_changes = app.last_component_state_changes
        assert switch in last_component_state_changes
        assert last_component_state_changes[switch].get("is_on") is True



import asyncio
import random
import re
import tarfile
import tempfile
from pathlib import Path
from typing import *  # type: ignore

import revel
from revel import input, print, print_chapter, warning

import rio.cli
from rio.cli.rio_api import RioApi


def should_directory_likely_be_excluded(
    dir_path: Path,
) -> tuple[str, str] | None:
    """
    Some directories should very likely not be part of the user's project. This
    function looks at a directory, and if the directory should likely be
    excluded it returns a name and explanation of why. Returns `None` otherwise.
    """
    assert dir_path.is_dir(), dir_path

    # Virtual environments
    if (dir_path / "pyvenv.cfg").exists():
        return (
            "contain a python virtual environment",
            "Virtual environments contain a huge number of files which are typically created by `pip`, `poetry`, `conda` or similar. These don't have to be included with your project and can easily be recreated later.",
        )

    # .git
    if dir_path.name == ".git":
        return (
            "be a git folder",
            "Git repositories contain the entire history of your project and typically aren't necessary to deploy your project.",
        )

    # node_modules
    if dir_path.name == "node_modules":
        return (
            "be npm's modules directory",
            "Node's package manager (npm) can store a vast number of files in these folders. These typically don't have to be included with your project and can easily be recreated later.",
        )

    # Build directories
    if dir_path.name in ("build", "dist"):
        return (
            "contain past build files",
            "Build directories can contain a large number of past build outputs and are not typically necessary to deploy your project.",
        )

    # All good, keep it
    return None


def list_files(
    proj: rio.cli.project.RioProject,
    dir_path: Path,
) -> Iterable[Path]:
    """
    Recursively yield all files in the given directory, ignoring any files that
    are ignored by the project.

    Any directories that should likely be excluded are ignored, unless
    explicitly listed in an exception in the `.rioignore` file.

    Interacts with the terminal.

    All yielded paths are absolute.
    """
    assert dir_path.exists(), dir_path
    assert dir_path.is_dir(), dir_path

    dir_path = dir_path.resolve()

    for path in dir_path.iterdir():
        # Ignore files that are ignored by the project
        if proj.ignores.is_path_ignored(path):
            continue

        # If this is a file, yield it
        if path.is_file():
            yield path
            continue

        # Is this a directory that the user likely doesn't want to include?
        exclude_reason = should_directory_likely_be_excluded(path)
        if (
            exclude_reason is not None
            and not proj.ignores.is_explicitly_included(path)
        ):
            appears_to, explanation = exclude_reason
            rel_path = path.relative_to(proj.project_directory)
            warning(
                f'Excluding "{rel_path}". This directory appears to {appears_to}.'
            )
            warning(explanation)
            warning(
                f'If you do want to include it after all, add the following to your ".rioignore" file:'
            )
            warning(f"!{rel_path}")
            print()

            # Explicitly ignore this directory
            proj.rioignore_additions.extend(
                [
                    f"# Automatically excluded by Rio",
                    f"# This directory appears to {appears_to}",
                    f"/{rel_path}",
                    "",
                ]
            )
            continue

        # Recurse
        yield from list_files(proj, path)


def pack_up_project(
    proj: rio.cli.project.RioProject,
    archive_path: Path,
) -> int:
    """
    Compresses all files in the project into an archive, displaying progress,
    and interacting with the terminal in general.

    Returns the size of the uncompressed files in bytes.
    """

    # Find all files which are part of the project
    print("Scanning project")
    project_directory = proj.project_directory.resolve()
    files = set(list_files(proj, project_directory))

    # Make sure essential files are included
    essential_files = {
        project_directory / "rio.toml",
    }

    missing_files = essential_files - files

    if missing_files:
        raise NotImplementedError(
            "TODO: Ask the user whether these files should be included"
        )

    # Gather all files into a single archive
    #
    # Randomize the file order to ensure that the progress is somewhat uniform.
    # Hard drives hate him.
    print("Creating app package")
    files = list(files)
    random.shuffle(files)
    total_size = 0

    with revel.ProgressBar() as bar:
        with tarfile.open(archive_path, "w:xz") as tar:
            for ii, file_path in enumerate(files):
                bar.progress = ii / len(files)

                # Add the file to the tarball
                relative_path = file_path.relative_to(project_directory)
                tar.add(file_path, arcname=relative_path)

                # Add the file size to the total
                total_size += file_path.stat().st_size

    return total_size


async def create_or_update_app(
    api: RioApi,
    proj: rio.cli.project.RioProject,
) -> None:
    """
    Uploads the given archive file to the cloud, creating a new deployment.
    """
    # Get an app name
    try:
        name = proj.deploy_name
    except KeyError:
        print(
            "What should your app be called? This name will be used as part of the URL."
        )
        print(
            'For example, if you name your app "my-app", it will be deployed at `https://rio.dev/.../my-app`.'
        )

        name = input("App name: ")

        allowed_chars = "abcdefghijklmnopqrstuvwxyz0123456789-"
        normalized = re.sub("[^" + allowed_chars + "]", "-", name.lower())
        normalized = re.sub("-+", "-", normalized)

    # Make sure the user is logged in
    # TODO

    # Pack up the project
    print_chapter("Packaging project")
    with tempfile.TemporaryDirectory() as tmp_dir:
        assert Path(tmp_dir).exists(), tmp_dir
        archive_path = Path(tmp_dir) / "packed-project.tar.xz"
        uncompressed_size_in_bytes = pack_up_project(proj, archive_path)
        compressed_size_in_bytes = archive_path.stat().st_size

        print(
            f"Compressed size: {compressed_size_in_bytes / 1024 / 1024:.2f} MiB"
        )
        print(
            f"Uncompressed size: {uncompressed_size_in_bytes / 1024 / 1024:.2f} MiB"
        )
        print(
            f"Compression ratio: {uncompressed_size_in_bytes / compressed_size_in_bytes:.2f}x"
        )

    # Make sure the user has the ability to create this app
    # TODO:
    # - # of apps
    # - compressed size
    # - uncompressed size

    # Create the app
    await api.create_app(
        name=app_name,
        packed_app=archive_path.open("rb"),
        realm=realm,
        start=start,
    )


async def main():
    with rio.cli.project.RioProject.try_locate_and_load() as proj:
        async with rio.cli.cli_instance.CliInstance() as cli:
            await create_or_update_app(None, proj)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from typing import Literal, final

from typing_extensions import Self

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["ListView"]


@final
class ListView(FundamentalComponent):
    """
    Vertically arranges and styles its children.

    A container which arranges its children in a vertical list. It is similar to `Column`,
    but it is optimized for displaying large numbers of items.


    ## Attributes

    `children`: The children to display in the list.

    `key`: A unique key for this component. If the key changes, the component will be
            destroyed and recreated. This is useful for components which maintain state
            across rebuilds.


    ## Examples

    A minimal example of a `ListView` with two items will be shown:

    ```python
    rio.ListView(
        rio.SimpleListItem("Product 1", key="item1"),
        rio.SimpleListItem("Product 2", key="item2"),
    )
    ```

    `ListView`s are commonly used to display lists of dynamic length. You can easily
    achieve this by adding the children to a list first, and then unpacking that list:

    ```python
    import functools


    class MyComponent(rio.Component):
        products: list[str] = ["Product 1", "Product 2", "Product 3"]

        def on_press_heading_list_item(self, product: str) -> None:
            print(f"Selected {product}")

        def build(self) -> rio.Component:
            # Store all children in an intermediate list
            list_items = []

            for product in self.products:
                list_items.append(
                    rio.SimpleListItem(
                        text=product,
                        key=product,
                        # Note the use of `functools.partial` to pass the product
                        # to the event handler.
                        on_press=functools.partial(
                            self.on_press_heading_list_item,
                            product=product,
                        ),
                    )
                )

            # Then unpack the list to pass the children to the `ListView`
            return rio.ListView(*list_items)
    ```
    """

    children: list[rio.Component]

    def __init__(
        self,
        *children: rio.Component,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ) -> None:
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)

    def add(self, child: rio.Component) -> Self:
        """
        Appends a child component.

        Appends a child component to the end and then returns the `ListView`,
        which makes method chaining possible:

        ```python
        rio.ListView().add(child1).add(child2)
        ```

        ## Parameters

        `child`: The child component to append.
        """
        self.children.append(child)
        return self


ListView._unique_id = "ListView-builtin"



import pytest

import imy.docstrings

import rio.docs


def _create_tests():
    for obj, docs in rio.docs.find_documented_objects(postprocess=True):
        if isinstance(docs, imy.docstrings.FunctionDocs):
            test_cls = _create_function_tests(docs)
        else:
            assert isinstance(obj, type)
            test_cls = _create_class_tests(obj, docs)

        globals()[test_cls.__name__] = test_cls


def _create_function_tests(docs: imy.docstrings.FunctionDocs) -> type:
    class Tests:  # type: ignore
        def test_summary(self) -> None:
            assert docs.summary is not None, f"{docs.name} has no summary"

        def test_details(self) -> None:
            assert docs.details is not None, f"{docs.name} has no details"

        @pytest.mark.parametrize(
            "param",
            docs.parameters,
            ids=[param.name for param in docs.parameters],
        )
        def test_param_description(
            self, param: imy.docstrings.FunctionParameter
        ) -> None:
            assert (
                param.description is not None
            ), f"{docs.name}.{param.name} has no description"

    Tests.__name__ = f"Test_{docs.name}"
    return Tests


def _create_class_tests(cls: type, docs: imy.docstrings.ClassDocs) -> type:
    methods = [
        func
        for func in docs.functions
        if func.name != "__init__" or not issubclass(cls, rio.Component)
    ]
    parametrize_methods = pytest.mark.parametrize(
        "method",
        methods,
        ids=[method.name for method in methods],
    )

    class Tests:
        def test_summary(self) -> None:
            assert docs.summary is not None, f"{cls.__name__} has no summary"

        def test_details(self) -> None:
            assert docs.details is not None, f"{cls.__name__} has no details"

        @pytest.mark.parametrize(
            "attr",
            docs.attributes,
            ids=[attr.name for attr in docs.attributes],
        )
        def test_attr_description(
            self, attr: imy.docstrings.ClassField
        ) -> None:
            assert (
                attr.description is not None
            ), f"{cls.__name__}.{attr.name} has no description"

        @parametrize_methods
        def test_method_summary(
            self, method: imy.docstrings.FunctionDocs
        ) -> None:
            assert (
                method.summary is not None
            ), f"{cls.__name__}.{method.name} has no summary"

        @parametrize_methods
        def test_method_details(
            self, method: imy.docstrings.FunctionDocs
        ) -> None:
            assert (
                method.details is not None
            ), f"{cls.__name__}.{method.name} has no details"

        @parametrize_methods
        def test_method_parameters(
            self, method: imy.docstrings.FunctionDocs
        ) -> None:
            for param in method.parameters[1:]:
                assert (
                    param.description is not None
                ), f"Parameter {param.name!r} has no description"

    Tests.__name__ = f"Test{docs.name}"
    return Tests


_create_tests()



# This module silences these annoying errors:
#
# ERROR:asyncio:Exception in callback _ProactorBasePipeTransport._call_connection_lost(None)
# handle: <Handle _ProactorBasePipeTransport._call_connection_lost(None)>
# Traceback (most recent call last):
#   File "C:\Python312\Lib\asyncio\events.py", line 84, in _run
#     self._context.run(self._callback, *self._args)
#   File "C:\Python312\Lib\asyncio\proactor_events.py", line 165, in _call_connection_lost
#     self._sock.shutdown(socket.SHUT_RDWR)
# ConnectionResetError: [WinError 10054] An existing connection was forcibly closed by the remote host

from asyncio.proactor_events import _ProactorBasePipeTransport

import introspection


class SocketWrapper:
    def __init__(self, socket):
        self._socket = socket

    def shutdown(self, *args, **kwargs):
        try:
            self._socket.shutdown(*args, **kwargs)
        except ConnectionResetError:
            pass

    def __getattr__(self, attr: str):
        return getattr(self._socket, attr)


def _call_connection_lost(
    wrapped_func, self: _ProactorBasePipeTransport, *args, **kwargs
):
    self._sock = SocketWrapper(self._sock)  # type: ignore

    # Call the original method
    try:
        wrapped_func(self, *args, **kwargs)
    finally:
        # Clean up
        if self._sock is not None:  # type: ignore
            self._sock = self._sock._socket  # type: ignore


introspection.wrap_method(_ProactorBasePipeTransport, _call_connection_lost)



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import final

from uniserde import JsonDoc

import rio

from .. import cursor_style
from ..color import Color
from .fundamental_component import FundamentalComponent

__all__ = [
    "Rectangle",
]


@final
class Rectangle(FundamentalComponent):
    """
    A customizable rectangle shape.

    Rectangles are versatile components that can be used to build up more
    complex elements. While not particularly interesting on their own, combining
    a rectangle with other components allows you to quickly create custom
    buttons, cards, or anything els you may need in your app.

    Rectangles also act as a simple source of animations. They have two styles:
    A default style for when the user isn't interacting with them, and a hover
    style for when the mouse hovers above them. This, along with their
    `transition_time` attribute allows you to make your app feel dynamic and
    alive.


    ## Attributes

    `content`: The component to display inside the rectangle.

    `fill`: The background color/image/gradient of the rectangle.

    `stroke_width`: The width of the rectangle's outline.

    `stroke_color`: The color of the rectangle's outline.

    `corner_radius`: The rectangle's corner radius. Can be a single number or a
        sequence of 4 numbers.

    `shadow_radius`: The corner radius of the rectangle's shadow.

    `shadow_offset_x`: The horizontal offset of the rectangle's shadow. A
        negative value moves the shadow to the left side of the rectangle.

    `shadow_offset_y`: The vertical offset of the rectangle's shadow. A
        negative value moves the shadow above the rectangle.

    `shadow_color`: The color of the rectangle's shadow.

    `hover_fill`: The rectangle's `fill` while the cursor is hovering over it.

    `hover_stroke_width`: The rectangle's `stroke_width` while the cursor is
        hovering over it.

    `hover_stroke_color`: The rectangle's `stroke_color` while the cursor is
        hovering over it.

    `hover_corner_radius`: The rectangle's `corner_radius` while the cursor is
        hovering over it.

    `hover_shadow_radius`: The rectangle's `shadow_radius` while the cursor is
        hovering over it.

    `hover_shadow_offset_x`: The rectangle's `shadow_offset_x` while the cursor
        is hovering over it.

    `hover_shadow_offset_y`: The rectangle's `shadow_offset_y` while the cursor
        is hovering over it.

    `hover_shadow_color`: The rectangle's `shadow_color` while the cursor is
        hovering over it.

    `transition_time`: How many seconds it should take for the rectangle to
        transition between its regular and hover styles.

    `cursor`: The cursor to display when the mouse hovers above the rectangle.

    `ripple`: Whether to display a Material Design ripple effect when the
        rectangle is hovered or clicked.


    ## Examples

    A minimal example of a rectangle with a text and red background will be
    shown:

    ```python
    rio.Rectangle(
        content=rio.Text("Hello World!"),
        fill=rio.Color.from_hex("ff0000"),
    )
    ```

    You can fill your `Rectangle` with an image instead of a color:

    ```python
    from pathlib import Path

    PATH = Path(__file__).parent

    rio.Rectangle(
        content=rio.Text("Hello World!"),
        fill=rio.ImageFill(
            PATH / "example_image.jpg",
            fill_mode="zoom",
        ),
    )
    ```
    """

    _: KW_ONLY
    content: rio.Component | None = None
    transition_time: float = 1.0
    cursor: rio.CursorStyle = cursor_style.CursorStyle.DEFAULT
    ripple: bool = False

    fill: rio.FillLike
    stroke_width: float = 0.0
    stroke_color: rio.Color = Color.BLACK
    corner_radius: float | tuple[float, float, float, float] = 0.0
    shadow_radius: float = 0.0
    shadow_offset_x: float = 0.0
    shadow_offset_y: float = 0.0
    shadow_color: rio.Color = Color.BLACK

    hover_fill: rio.FillLike | None = None
    hover_stroke_width: float | None = None
    hover_stroke_color: rio.Color | None = None
    hover_corner_radius: float | tuple[float, float, float, float] | None = None
    hover_shadow_radius: float | None = None
    hover_shadow_offset_x: float | None = None
    hover_shadow_offset_y: float | None = None
    hover_shadow_color: rio.Color | None = None

    def __post_init__(self):
        self.fill = rio.Fill._try_from(self.fill)

    def _custom_serialize(self) -> JsonDoc:
        return {
            # Regular
            "fill": rio.Fill._try_from(self.fill)._serialize(self._session_),
            "corner_radius": (
                self.corner_radius
                if self.corner_radius is None
                or isinstance(self.corner_radius, tuple)
                else (self.corner_radius,) * 4
            ),
            # Hover
            "hover_fill": (
                None
                if self.hover_fill is None
                else rio.Fill._try_from(self.hover_fill)._serialize(
                    self._session_
                )
            ),
            "hover_corner_radius": (
                self.hover_corner_radius
                if self.hover_corner_radius is None
                or isinstance(self.hover_corner_radius, tuple)
                else (self.hover_corner_radius,) * 4
            ),
        }


Rectangle._unique_id = "Rectangle-builtin"



from pathlib import Path

import rio

# Initialize the development component
HERE = Path(__file__).parent.resolve()

rio.DevelComponent.initialize(HERE)


app = rio.App(
    build=lambda: rio.DevelComponent(
        children=[
            rio.Column(
                rio.NodeInput("input 1", rio.Color.RED, key="input1"),
                rio.Text("Child 1"),
                rio.NodeOutput("output 1", rio.Color.RED, key="output1"),
            ),
            rio.NodeInput("input 2", rio.Color.RED, key="input2"),
            rio.NodeOutput("output 2", rio.Color.RED, key="output2"),
        ],
    ),
    theme=rio.Theme.from_colors(light=False),
)

app.run_as_web_server(
    port=8001,
)



from utils import create_mockapp

import rio


async def test_fundamental_container_as_root():
    def build():
        return rio.Row(rio.Text("Hello"))

    async with create_mockapp(build) as app:
        row_component = app.get_component(rio.Row)
        text_component = app.get_component(rio.Text)

        assert {row_component, text_component} <= app.last_updated_components



__version__ = "0.6.4"

import logging

_logger = logging.getLogger(__name__)

# Re-export dataclass stuff for easy use.
from dataclasses import KW_ONLY as KW_ONLY
from dataclasses import field as field

# URLs are used as an important datatype within rio. Re-export them for easy
# use.
from yarl import URL as URL

from . import event as event
from . import patches_for_3rd_party_stuff
from .app import *
from .color import *
from .common import *
from .components import *
from .cursor_style import *
from .errors import *
from .fills import *
from .routing import *
from .session import *
from .text_style import *
from .theme import *
from .user_settings_module import *

del patches_for_3rd_party_stuff



import enum

__all__ = ["CursorStyle"]


class CursorStyle(enum.Enum):
    """
    Enumeration of all available cursor styles. Use these to indicate which
    kinds of action a user can perform while hovering over a component.
    """

    # TODO: Extend these
    DEFAULT = enum.auto()
    POINTER = enum.auto()



from typing import *  # type: ignore

import rio


class LayoutPreview(rio.Component):
    """
    WIP component for previewing the layout of another component in the
    debugger.
    """

    component: rio.Component

    def build(self) -> rio.Component:
        palette = self.session.theme.neutral_palette

        return rio.Rectangle(
            content=rio.Rectangle(
                margin=self.component.margin,
                margin_x=self.component.margin_x,
                margin_y=self.component.margin_y,
                margin_left=self.component.margin_left,
                margin_top=self.component.margin_top,
                margin_right=self.component.margin_right,
                margin_bottom=self.component.margin_bottom,
                width=self.component.width,
                height=self.component.height,
                align_x=self.component.align_x,
                align_y=self.component.align_y,
                fill=palette.background_active,
            ),
            fill=palette.background_variant,
        )



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Literal, final

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Drawer",
    "DrawerOpenOrCloseEvent",
]


@final
@dataclass
class DrawerOpenOrCloseEvent:
    """
    The argument for the `on_open_or_close` handler of a `Drawer`.

    ## Attributes

    `is_open`: The new `is_open` state of the `Drawer`.
    """

    is_open: bool


@final
class Drawer(FundamentalComponent):
    """
    A container which slides in from the edge of the screen.

    Drawers are containers which can either be completely hidden from view, or
    be made visible by sliding in from the edge of the screen. They are commonly
    used for navigation on smaller displays.

    Drawers take two children: The `anchor` is always visible and positions the
    drawer. The `content` is located inside the drawer and is only visible when
    the drawer is open.

    Drawers have the ability to be `modal`. Modal drawers draw attention to
    themselves and prevent interaction with the anchor while open.


    ## Attributes

    `anchor`: A component which is always visible and positions the drawer.

    `content`: A component which is only visible when the drawer is open.

    `on_open_or_close`: Triggered whenever the user opens or closes the
        drawer.

    `side`: The side of the screen from which the drawer slides in.

    `is_modal`: Whether the drawer should prevent interaction with the anchor
        while open.

    `is_open`: Whether the drawer is currently open.

    `is_user_openable`: Whether the user can open or close the drawer. If this
        is `False`, the drawer can only be opened or closed
        programmatically.


    ## Examples

    A simple drawer with a button as the anchor and some text as content:

    ```python
    rio.Drawer(
        anchor=rio.Button("Click Me!"),
        content=rio.Text("It was clickbait!"),
    )
    ```

    The drawer has a button as its anchor and a switcher bar as its content.
    The functionality is such that when the button is clicked, the drawer will
    toggle between open and closed states:

    ```python
    class MyComponent(rio.Component):
        is_open: bool = False

        def on_press_button(self) -> None:
            self.is_open = True

        def build(self) -> rio.Component:
            return rio.Drawer(
                anchor=rio.Button(
                    "Click Me!",
                    on_press=self.on_press_button,
                ),
                content=rio.Text(
                    "It was clickbait!",
                ),
                is_open=self.is_open,
            )
    ```
    """

    anchor: rio.Component
    content: rio.Component
    _: KW_ONLY
    on_open_or_close: rio.EventHandler[DrawerOpenOrCloseEvent] = None
    side: Literal["left", "right", "top", "bottom"] = "left"
    is_modal: bool = True
    is_open: bool = False
    is_user_openable: bool = True

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"is_open"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "is_open" in delta_state and not self.is_user_openable:
            raise AssertionError(
                "Frontend tried to change value of `Drawer.is_open` even though `is_user_openable` is `False`"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_open_or_close event
        try:
            is_open = delta_state["is_open"]
        except KeyError:
            pass
        else:
            assert isinstance(is_open, bool), (is_open, type(is_open))
            await self.call_event_handler(
                self.on_open_or_close,
                DrawerOpenOrCloseEvent(is_open),
            )


Drawer._unique_id = "Drawer-builtin"



from typing import *  # type:ignore

import pandas as pd


# example data for our bar chart in the balance card component
BAR_CHART: pd.DataFrame = pd.DataFrame([4, 5, 6, 5, 4, 6, 7], columns=["data"])


# example data for our crypto portfolio. Our portfolio consists of three coins:
# bitcoin, litecoin, and ethereum. Each coin has a value, a ticker, a color, and a logo.
MY_COINS: dict[str, Tuple[float, str, str, str]] = {
    "bitcoin": (
        13.344546,
        "BTC",
        "#f7931a",
        "https://cryptologos.cc/logos/bitcoin-btc-logo.svg?v=029",
    ),
    "litecoin": (
        40.21321,
        "LTC",
        "#365d99",
        "https://cryptologos.cc/logos/litecoin-ltc-logo.svg?v=029",
    ),
    "ethereum": (
        4.239234,
        "ETH",
        "#14044d",
        "https://cryptologos.cc/logos/ethereum-eth-logo.svg?v=029",
    ),
}



import functools
import html
import os
import sys
import traceback
import types
from pathlib import Path
from typing import *  # type: ignore

import revel

import rio
import rio.icon_registry

from .. import nice_traceback, project


class AppLoadError(Exception):
    pass


def traceback_frame_filter(frame: traceback.FrameSummary) -> bool:
    # Skip frames which are internal to rio
    rio_root = rio.__file__
    assert rio_root.endswith(os.sep + "__init__.py")
    rio_root = rio_root.removesuffix(os.sep + "__init__.py")

    return not frame.filename.startswith(rio_root)


def make_traceback_html(
    *,
    err: Union[str, BaseException],
    project_directory: Path,
) -> str:
    icon_registry = rio.icon_registry.IconRegistry.get_singleton()
    error_icon_svg = icon_registry.get_icon_svg("material/error")

    if isinstance(err, str):
        traceback_html = html.escape(err)
    else:
        traceback_html = nice_traceback.format_exception_html(
            err,
            relpath=project_directory,
            frame_filter=traceback_frame_filter,
        )

    return f"""
<div>
    <div class="rio-traceback rio-debugger-background rio-switcheroo-neutral">
        <div class="rio-traceback-header">
            {error_icon_svg}
            <div>Couldn't load the app</div>
        </div>
        <div class="rio-traceback-message">
            Fix the issue below. The app will automatically reload once you save the
            file.
        </div>
        <div class="rio-traceback-traceback">{traceback_html}</div>
        <div class="rio-traceback-footer">
            Need help?
            <div class="rio-traceback-footer-links">
                <a class="rio-text-link" target="_blank" href="https://discord.gg/7ejXaPwhyH">Ask on Rio's Discord</a>
                <a class="rio-text-link" target="_blank" href="https://chat.openai.com">Ask ChatGPT</a>
                <a class="rio-text-link" target="_blank" href="https://rio.dev/docs">Read the docs</a>
            </div>
        </div>
    </div>
</div>
"""


def make_error_message_component(
    err: Union[str, BaseException],
    project_directory: Path,
) -> rio.Component:
    html = make_traceback_html(
        err=err,
        project_directory=project_directory,
    )

    return rio.Html(html)


def make_error_message_app(
    err: Union[str, BaseException],
    project_directory: Path,
    theme: rio.Theme | Tuple[rio.Theme, rio.Theme],
) -> rio.App:
    """
    Creates an app that displays the given error message.
    """
    return rio.App(
        build=functools.partial(
            make_error_message_component, err, project_directory
        ),
        theme=theme,
    )


def import_app_module(proj: project.RioProject) -> types.ModuleType:
    """
    Python's importing is bizarre. This function tries to hide all of that and
    imports the module, as specified by the user. This can raise a variety of
    exceptions, since the module's code is evaluated.
    """
    # Purge the module from the module cache
    app_main_module = proj.app_main_module
    root_module, _, _ = app_main_module.partition(".")

    for module_name in list(sys.modules):
        if module_name.partition(".")[0] == root_module:
            del sys.modules[module_name]

    # Now (re-)import the app module
    main_module_path = str(proj.module_path.parent)
    sys.path.append(main_module_path)

    try:
        return __import__(app_main_module)
    finally:
        sys.path.remove(main_module_path)


def load_user_app(proj: project.RioProject) -> rio.App:
    """
    Load and return the user app. Raises `AppLoadError` if the app can't be
    loaded for whichever reason.
    """
    # Import the app module
    try:
        app_module = import_app_module(proj)
    except BaseException as err:
        revel.error(f"Could not import `{proj.app_main_module}`:")

        revel.print(
            nice_traceback.format_exception_revel(
                err,
                relpath=proj.project_directory,
                frame_filter=traceback_frame_filter,
            )
        )

        raise AppLoadError() from err

    # Find the variable holding the Rio app
    apps: list[tuple[str, rio.App]] = []
    for var_name, var in app_module.__dict__.items():
        if isinstance(var, rio.App):
            apps.append((var_name, var))

    if app_module.__file__ is None:
        main_file_reference = f"Your app's main file"
    else:
        main_file_reference = f"The file `{Path(app_module.__file__).relative_to(proj.project_directory)}`"

    if len(apps) == 0:
        raise AppLoadError(
            f"Cannot find your app. {main_file_reference} needs to to define a variable that is a Rio app. Something like `app = rio.App(...)`"
        )

    if len(apps) > 1:
        variables_string = (
            "`" + "`, `".join(var_name for var_name, _ in apps) + "`"
        )
        raise AppLoadError(
            f"{main_file_reference} defines multiple Rio apps: {variables_string}. Please make sure there is exactly one."
        )

    app = apps[0][1]

    # Explicitly set the asset directory because it can't reliably be
    # auto-detected
    module_path = proj.module_path
    if module_path.is_file():
        module_path = module_path.parent
    app.assets_dir = module_path / app._assets_dir

    # If runnWrap the app's `build` function so that it displays a nice traceback
    # in case of an error

    return app
